/**
 * variant.hpp
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * https://gist.github.com/tibordp/6909880
 *
 * Created on: Feb 28, 2014
 * Author: Tibor Djurica Potpara
 * Modified: Jan Travnicek
 */

#ifndef __VATIANT_HPP_
#define __VATIANT_HPP_

#include <iostream>
#include <utility>
#include <typeinfo>
#include <cstring>
#include <sstream>

#include <extensions/compare.hpp>
#include <extensions/utility.hpp>
#include <extensions/type_traits.hpp>
#include <extensions/typeindex.h>
#include <extensions/algorithm.hpp>

#include "string.hpp"

namespace ext {

/**
 * \brief
 * Class designed to implement core features of varant.
 *
 * \tparam Ts ... pack of types of variant
 */
template<typename... Ts>
struct variant_helper;

/**
 * \brief
 * Class designed to implement core of variant features for individual non-void types.
 *
 * \tparam F the currently handled type
 * \tparam Ts ... pack of remaining types to handle
 */
template<typename F, typename... Ts>
struct variant_helper<F, Ts...> {
	/**
	 * \brief
	 * Implementation of call passing for type F (if the type F is currently stored inside the variant).
	 *
	 * \tparam Result the callback result
	 * \tparam Callable type of callback to execute
	 *
	 * \param id the representation of type stored inside the variant
	 * \param data storage occupied by value stored in the variant
	 * \param callable the callback to execute
	 */
	template < class Result, class Callable >
	inline static Result call ( ext::type_index id, const void * data, Callable callable ) {
		if (id == ext::type_index ( typeid ( F ) ) )
			return callable ( * reinterpret_cast < const F * > ( data ) );
		else
			return variant_helper < Ts ... >::template call < Result > ( id, data, callable );
	}

	/**
	 * \brief
	 * Implementation of destructor for type F (if the type F is currently stored inside the variant)
	 *
	 * \param id the representation of type stored inside the variant
	 * \param data storage occupied by value stored in the variant
	 */
	inline static void destroy(ext::type_index id, void * data) {
		if (id == ext::type_index ( typeid ( F ) ) )
			reinterpret_cast<F*>(data)->~F();
		else
			variant_helper<Ts...>::destroy(id, data);
	}

	/**
	 * \brief
	 * Implementation of copy constructor for type F (if the type F is currently stored inside the variant)
	 *
	 * \param id the representation of type stored inside the variant
	 * \param old_v storage occupied by value stored in the source variant
	 * \param new_v storage of the target variant
	 */
	inline static void copy(ext::type_index old_t, const void * old_v, void * new_v) {
		if (old_t == ext::type_index ( typeid ( F ) ))
			new (new_v) F(*reinterpret_cast<const F*>(old_v));
		else
			variant_helper<Ts...>::copy(old_t, old_v, new_v);
	}

	/**
	 * \brief
	 * Implementation of move constructor for type F (if the type F is currently stored inside the variant)
	 *
	 * \param id the representation of type stored inside the variant
	 * \param old_v storage occupied by value stored in the source variant
	 * \param new_v storage of the target variant
	 */
	inline static void move(ext::type_index old_t, void * old_v, void * new_v) {
		if (old_t == ext::type_index ( typeid ( F ) ))
			new (new_v) F(std::move(*reinterpret_cast<F*>(old_v)));
		else
			variant_helper<Ts...>::move(old_t, old_v, new_v);
	}

	/**
	 * \brief
	 * Implementation of three way comparison for type F (if the type F is currently stored inside the variant)
	 *
	 * \param this_t the representation of type stored inside the first variant
	 * \param this_v storage of the first variant
	 * \param other_t the representation of type stored inside the second variant
	 * \param other_v storage of the second variant
	 *
	 * \return the result of the comparison
	 */
	inline static int compareHelper(ext::type_index this_t, const void * this_v, ext::type_index other_t, const void * other_v) {
		if (this_t == ext::type_index ( typeid ( F ) ) && other_t == ext::type_index ( typeid ( F ) )) {
			static ext::compare<typename std::decay < F >::type > comp;
			return comp( *(reinterpret_cast<const F*>(this_v)), *(reinterpret_cast<const F*>(other_v)));
		} else
			return variant_helper<Ts...>::compareHelper(this_t, this_v, other_t, other_v);
	}
};

/**
 * \brief
 * Class ending the recursive handling of possible types in the variant type.
 */
template < > struct variant_helper < > {

	/**
	 * \brief
	 * Implementation of call passing for type F (if the type F is currently stored inside the variant).
	 *
	 * Does nothing.
	 */
	template < class Result, class Callable >
	inline static Result call ( ext::type_index, const void *, Callable ) {
		throw std::logic_error ( "Out of types of variant." );
	}

	/**
	 * \brief
	 * Destructor helper when the pack of types of the variant was depleted.
	 *
	 * Does nothing.
	 */
	inline static void destroy(ext::type_index, void *) { }

	/**
	 * \brief
	 * Copy constructor helper when the pack of types of the variant was depleted.
	 *
	 * Does nothing.
	 */
	inline static void copy(ext::type_index, const void *, void *) { }

	/**
	 * \brief
	 * Move constructor helper when the pack of types of the variant was depleted.
	 *
	 * Does nothing.
	 */
	inline static void move(ext::type_index, void *, void *) {  }

	/**
	 * \brief
	 * Comparison helper when the pack of types of the variant was depleted.
	 *
	 * Does nothing.
	 *
	 * \return 0
	 */
	inline static int compareHelper(ext::type_index, const void *, ext::type_index, const void *) { return 0; }
};

/**
 * \brief
 * Class representing functionality of the variant specific for individual types of variant.
 *
 * \tparam ST the overal maximum size of types to be stored inside the variant
 * \tparam AT the overal maximal alignment of types to be stored inside the variant
 *
 * \tparam Ts ... pack of variant types
 */
template<size_t ST, size_t AT, typename... Ts>
class variant_base;


/**
 * \brief
 * Class ending the recursive construction of variant_base type. The class is responsible for actual allocation of storage of size ST with alignment AT
 *
 * \tparam ST the overal maximum size of types to be stored inside the variant
 * \tparam AT the overal maximal alignment of types to be stored inside the variant
 */
template<size_t ST, size_t AT>
class variant_base<ST, AT> {
	/**
	 * \brief
	 * The type of storage capable of containing any of the concrete variant type.
	 */
	using data_t = typename std::aligned_storage < ST, AT>::type;

protected:
	/**
	 * \brief
	 * Representation of current type stored inside the variant.
	 */
	ext::type_index type_id;

	/**
	 * \brief
	 * Representation of the value stored inside the variant.
	 */
	data_t data;

	/**
	 * \brief
	 * Helper constructor setting up the type stored inside the variant
	 *
	 * \param id the identification of type stored inside the variant
	 */
	explicit variant_base( ext::type_index id ) : type_id ( id ) {
		// just to make the -Werror=maybe-uninitialized go away
		std::memset( & data, 0, ST );
	}
};

/**
 * \brief
 * Class representing functionality of the variant specific for selected non-void type of variant.
 *
 * \tparam ST the overal maximum size of types to be stored inside the variant
 * \tparam AT the overal maximal alignment of types to be stored inside the variant
 *
 * \tparam F the currently handled type
 * \tparam Ts ... pack of variant types
 */
template<size_t ST, size_t AT, typename F, typename... Ts>
class variant_base<ST, AT, F, Ts...> : public variant_base<ST, AT, Ts...> {
	/**
	 * \brief
	 * Use constructors of the base class
	 */
	using variant_base<ST, AT, Ts...>::variant_base;

protected:
	/**
	 * \brief
	 * Pass the construction of the variant with predefined type specification lower in the type recursion
	 *
	 * \param id the identification of type stored inside the variant
	 */
	explicit variant_base( ext::type_index id ) : variant_base<ST, AT, Ts...>::variant_base ( id ) { }

public:
	/**
	 * \brief
	 * Alows construction of the variant based on const referenced value of type F.
	 *
	 * \param value the value to base the variant content on
	 */
	variant_base(const F& value) : variant_base<ST, AT, Ts...>::variant_base ( ext::type_index ( typeid(F) ) ) { new (&this->data) F(value); }

	/**
	 * \brief
	 * Alows construction of the variant based on rvalue referenced value of type F.
	 *
	 * \param value the value to base the variant content on
	 */
	variant_base(F&& value) : variant_base<ST, AT, Ts...>::variant_base ( ext::type_index ( typeid(F) ) ) { new (&this->data) F(std::move(value)); }

	/**
	 * \brief
	 * Pass maybe exitent default constructor of the variant.
	 */
	variant_base() : variant_base<ST, AT, Ts ... >::variant_base ( ) { }
};

/**
 * \brief
 * Implementation of the variant class allowing to store any type of those listed in the template parameters. Void type is allowed, multiply specified same type is irrelevant.
 *
 * \tparam Ts ... pack of allowed types.
 */
template<typename ... Ts>
class variant : public variant_base < max ( SizeOf < Ts >::size ... ), max ( AlignOf < Ts >::align ... ), Ts...> {
	/**
	 * \brief
	 * Type of the helper used for type specific operations on the variant.
	 */
	using helper_t = variant_helper<Ts...>;

	/**
	 * Internal variant to string callback implementation.
	 */
	class VariantToString {
	public:
		/**
		 * The callback for types different from void
		 *
		 * \tparam T the accepted tye
		 *
		 * \param value the accepted value
		 *
		 * \return string representation of the value
		 */
		template < class T >
		std::string operator ( ) ( const T & value ) const {
			return ext::to_string ( value );
		}
	};

	/**
	 * Internal variant to ostream callback implementation.
	 */
	class VariantToStream {
		/**
		 * The stream to print to.
		 */
		std::ostream & m_out;
	public:
		/**
		 * Constructor of the variant to stream callback.
		 *
		 * \param out the output stream to use in function call operators.
		 */
		explicit VariantToStream ( std::ostream & out ) : m_out ( out ) {
		}

		/**
		 * The callback for types different from void
		 *
		 * \tparam T the accepted tye
		 *
		 * \param value the accepted value
		 */
		template < class T >
		void operator ( ) ( const T & value ) {
			m_out << value;
		}
	};
public:
	/**
	 * \brief
	 * Inherit type specific constructors of the variant.
	 */
	using variant_base < max ( SizeOf < Ts >::size ... ), max ( AlignOf < Ts >::align ... ), Ts...>::variant_base;

	/**
	 * \brief
	 * Default constructor allowed if the void type is in the pack of variant's possible type.
	 */
	variant ( ) : variant_base < max ( SizeOf < Ts >::size ... ), max ( AlignOf < Ts >::align ... ), Ts...> ( ) {
	}

	/**
	 * \brief
	 * Copy constructor of the variant.
	 *
	 * \param old the source instance
	 */
	variant ( const variant < Ts ... > & old ) : variant_base < max ( SizeOf < Ts >::size ... ), max ( AlignOf < Ts >::align ... ), Ts...> ( old.type_id ) {
		helper_t::copy(old.type_id, &old.data, &this->data);
	}

	/**
	 * \brief
	 * Move constructor of the variant.
	 *
	 * \param old the source instance
	 */
	variant ( variant < Ts ... > && old ) noexcept : variant_base < max ( SizeOf < Ts >::size ... ), max ( AlignOf < Ts >::align ... ), Ts...> ( old.type_id ) {
		helper_t::move(old.type_id, &old.data, &this->data);
	}

	/**
	 * \brief
	 * Copy assignment operator of the variant.
	 *
	 * \param old the source instance
	 *
	 * \return the modified target instance
	 */
	variant < Ts ... > & operator = ( const variant < Ts ... > & other ) {
		return this->operator =( variant < Ts ... > ( other ) );
	}

	/**
	 * \brief
	 * Move assignment operator of the variant.
	 *
	 * \param old the source instance
	 *
	 * \return the modified target instance
	 */
	variant < Ts ... > & operator = ( variant < Ts ... > && old ) noexcept {
		helper_t::destroy(this->type_id, &this->data);
		helper_t::move(old.type_id, &old.data, &this->data);
		this->type_id = old.type_id;

		return *this;
	}

	/**
	 * \brief
	 * Specialisation of equality operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if compared values are the same, false othervise
	 */
	bool operator == ( const variant < Ts ... > & other ) const {
		return compare ( other ) == 0;
	}

	/**
	 * \brief
	 * Specialisation of non-equality operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if compared values are not the same, false othervise
	 */
	bool operator != ( const variant < Ts ... > & other ) const {
		return !(*this == other);
	}

	/**
	 * \brief
	 * Specialisation of less than operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is smaller than the other instance, false othervise
	 */
	bool operator < ( const variant < Ts ... > & other ) const {
		return compare ( other ) < 0;
	}

	/**
	 * \brief
	 * Specialisation of more than operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is greater than the other instance, false othervise
	 */
	bool operator > ( const variant < Ts ... > & other ) const {
		return other < *this;
	}

	/**
	 * \brief
	 * Specialisation of less than or equal operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is smaller than or the same as the other instance, false othervise
	 */
	bool operator <= ( const variant < Ts ... > & other ) const {
		return !(*this > other);
	}

	/**
	 * \brief
	 * Specialisation of greater than or equal operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is greater than or the same as the other instance, false othervise
	 */
	bool operator >= ( const variant < Ts ... > & other ) const {
		return !(*this < other);
	}

	/**
	 * \brief
	 * Implementation of the three way comparison for varant.
	 *
	 * \param other the value compared to this
	 *
	 * \return negative value if this is smaller than the other instance, positive if this is greater than the other instance, zero othervise
	 */
	int compare ( const variant < Ts ... > & other ) const {
		if (this->type_id < other.type_id) return -1;
		if (this->type_id > other.type_id) return 1;

		return helper_t::compareHelper(this->type_id, &this->data, other.type_id, &other.data);
	}

	/**
	 * \brief
	 * Implementation of the three way comparison for varant.
	 *
	 * \tparam T the type of the other compared value
	 *
	 * \param other the value compared to this
	 *
	 * \return negative value if this is smaller than the other instance, positive if this is greater than the other instance, zero othervise
	 */
	template < typename T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
	int compare ( const T & other ) const {
		if ( this->type_id < ext::type_index ( typeid ( T ) ) ) return -1;
		if ( this->type_id > ext::type_index ( typeid ( T ) ) ) return 1;

		static ext::compare < typename std::decay < T >::type > comp;
		return comp ( this->template get < T > ( ), other );
	}

	/**
	 * \brief
	 * Function to test whether the variant currently contains type T.
	 *
	 * \tparam T the type to test
	 *
	 * \return true it the variant contains instance of type T
	 */
	template<typename T, typename std::enable_if < ext::is_in<T, Ts...>::value >::type* = nullptr >
	bool is() const {
		return (this->type_id == ext::type_index ( typeid ( T ) ));
	}

	/**
	 * \brief
	 * Allows to assign a value to the variant.
	 *
	 * \tparam T the type of the value to assign
	 *
	 * \param value the new value to assign to the variant
	 */
	template < typename T >
	void set ( typename std::enable_if < ext::is_in<T, Ts...>::value, T >::type && value ) {
		helper_t::destroy(this->type_id, &this->data);
		new ( & this->data ) T ( std::move ( value ) );
		this->type_id = ext::type_index ( typeid ( T ) );
	}

	/**
	 * \brief
	 * Allows to assign a value to the variant.
	 *
	 * \tparam T the type of the value to assign
	 *
	 * \param value the new value to assign to the variant
	 */
	template < typename T >
	void set ( const typename std::enable_if < ext::is_in<T, Ts...>::value, T >::type & value ) {
		helper_t::destroy(this->type_id, &this->data);
		new ( & this->data ) T ( value );
		this->type_id = ext::type_index ( typeid ( T ) );
	}

	/**
	 * \brief
	 * Allows to retrieve a value from the variant.
	 *
	 * \tparam T the type of the value to retrieve
	 *
	 * \return value retrieved from the variant
	 *
	 * \throws bad_cast exception if the variant stores different type than specified
	 */
	template < typename T >
	typename std::enable_if < ext::is_in<T, Ts...>::value, T >::type & get ( ) & {
		// It is a dynamic_cast-like behaviour
		if ( this->type_id == ext::type_index ( typeid ( T ) ) )
			return * reinterpret_cast < T * > ( & this->data );
		else
			throw std::bad_cast();
	}

	/**
	 * \brief
	 * Allows to retrieve a value from the variant.
	 *
	 * \tparam T the type of the value to retrieve
	 *
	 * \return value retrieved from the variant
	 *
	 * \throws bad_cast exception if the variant stores different type than specified
	 */
	template<typename T >
	const typename std::enable_if < ext::is_in<T, Ts...>::value, T >::type && get ( ) const & {
		// It is a dynamic_cast-like behaviour
		if ( this->type_id == ext::type_index ( typeid ( T ) ) )
			return std::move ( * reinterpret_cast < const T * > ( & this->data ) );
		else
			throw std::bad_cast();
	}

	/**
	 * \brief
	 * Allows to retrieve a value from the variant.
	 *
	 * \tparam T the type of the value to retrieve
	 *
	 * \return value retrieved from the variant
	 *
	 * \throws bad_cast exception if the variant stores different type than specified
	 */
	template < typename T >
	typename std::enable_if < ext::is_in<T, Ts...>::value, T >::type && get ( ) && {
		return std::move ( this->get < T > ( ) );
	}

	/**
	 * \brief
	 * Destructor of the variant.
	 */
	~variant ( ) noexcept {
		helper_t::destroy(this->type_id, &this->data);
	}

	/**
	 * \brief
	 * Callback executor on current variant value.
	 *
	 * \tparam Result the uniform result type
	 * \tparam Callable type of callable able to accept any variant type.
	 *
	 * \param callable object of callable type able to accept any variant type via function call operator. The possible void type is represented by call with no parameters
	 *
	 * \return the result of the callable on the current variable type
	 */
	template < class Result, class Callable >
	Result call ( Callable callable ) const {
		return helper_t::template call < Result > ( this->type_id, & this->data, callable );
	}

	/**
	 * \brief
	 * Operator to print the variant to the output stream.
	 *
	 * \param out the output stream
	 * \param obj the variant to print
	 *
	 * \return the output stream from the \p out
	 */
	friend std::ostream & operator << ( std::ostream & out, const variant < Ts ... > & obj ) {
		obj.call < void > ( VariantToStream ( out ) );
		return out;
	}

	/**
	 * \brief
	 * To string cast operator.
	 *
	 * \return string representation of the variant
	 */
	explicit operator std::string() const {
		return this->call < std::string > ( VariantToString ( ) );
	}

	/**
	 * \brief
	 * Allows to create the variant from defaultly constructed value of specified type.
	 *
	 * \tparam T the type to default construct ańd assign to newly constructed variant
	 *
	 * \return new variant with content of defaultly constructed value of type T
	 */
	template < typename T >
	static typename std::enable_if < ! std::is_same < void, T >::value, variant < Ts ... > >::type from ( ) {
		return variant < Ts ... > ( T { } );
	}
};

/**
 * \brief
 * Implementation of the variant class allowing to store any type of those listed in the template parameters. Void type is allowed, multiply specified same type is irrelevant.
 *
 * \tparam Ts ... pack of allowed types.
 */
template < typename T >
class variant < T > {
	T m_value;

public:
	/**
	 * \brief
	 * Alows construction of the variant based on const referenced value of type F.
	 *
	 * \param value the value to base the variant content on
	 */
	variant ( const T & value ) : m_value ( value ) {
	}

	/**
	 * \brief
	 * Alows construction of the variant based on rvalue referenced value of type F.
	 *
	 * \param value the value to base the variant content on
	 */
	variant ( T && value ) : m_value ( std::move ( value ) ) {
	}

	/**
	 * \brief
	 * Specialisation of equality operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if compared values are the same, false othervise
	 */
	bool operator == ( const variant < T > & other ) const {
		return compare ( other ) == 0;
	}

	/**
	 * \brief
	 * Specialisation of non-equality operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if compared values are not the same, false othervise
	 */
	bool operator != ( const variant < T > & other ) const {
		return !(*this == other);
	}

	/**
	 * \brief
	 * Specialisation of less than operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is smaller than the other instance, false othervise
	 */
	bool operator < ( const variant < T > & other ) const {
		return compare ( other ) < 0;
	}

	/**
	 * \brief
	 * Specialisation of more than operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is greater than the other instance, false othervise
	 */
	bool operator > ( const variant < T > & other ) const {
		return other < *this;
	}

	/**
	 * \brief
	 * Specialisation of less than or equal operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is smaller than or the same as the other instance, false othervise
	 */
	bool operator <= ( const variant < T > & other ) const {
		return !(*this > other);
	}

	/**
	 * \brief
	 * Specialisation of greater than or equal operator for varant.
	 *
	 * \param other the value compared to compared this
	 *
	 * \return true if this is greater than or the same as the other instance, false othervise
	 */
	bool operator >= ( const variant < T > & other ) const {
		return !(*this < other);
	}

	/**
	 * \brief
	 * Implementation of the three way comparison for varant.
	 *
	 * \param other the value compared to this
	 *
	 * \return negative value if this is smaller than the other instance, positive if this is greater than the other instance, zero othervise
	 */
	int compare ( const variant < T > & other ) const {
		static ext::compare < typename std::decay < T >::type > comp;
		return comp ( m_value, other.m_value );
	}

	/**
	 * \brief
	 * Implementation of the three way comparison for varant.
	 *
	 * \param other the value compared to this
	 *
	 * \return negative value if this is smaller than the other instance, positive if this is greater than the other instance, zero othervise
	 */
	int compare ( const T & other ) const {
		static ext::compare < typename std::decay < T >::type > comp;
		return comp ( m_value, other );
	}

	/**
	 * \brief
	 * Function to test whether the variant currently contains type T.
	 *
	 * \tparam T the type to test
	 *
	 * \return true it the variant contains instance of type T
	 */
	template < typename F, typename std::enable_if < std::is_same < F, T >::value >::type * = nullptr >
	bool is ( ) const {
		return true;
	}

	/**
	 * \brief
	 * Allows to assign a value to the variant.
	 *
	 * \tparam T the type of the value to assign
	 *
	 * \param value the new value to assign to the variant
	 */
	template < typename F, typename std::enable_if < std::is_same < F, T >::value >::type * = nullptr >
	void set ( F && value ) {
		m_value = std::move ( value );
	}

	/**
	 * \brief
	 * Allows to assign a value to the variant.
	 *
	 * \tparam T the type of the value to assign
	 *
	 * \param value the new value to assign to the variant
	 */
	template < typename F, typename std::enable_if < std::is_same < F, T >::value >::type * = nullptr >
	void set ( const F & value ) {
		m_value = value;
	}

	/**
	 * \brief
	 * Allows to retrieve a value from the variant.
	 *
	 * \tparam T the type of the value to retrieve
	 *
	 * \return value retrieved from the variant
	 *
	 * \throws bad_cast exception if the variant stores different type than specified
	 */
	template < typename F, typename std::enable_if < std::is_same < F, T >::value >::type * = nullptr >
	F & get ( ) & {
		return m_value;
	}

	/**
	 * \brief
	 * Allows to retrieve a value from the variant.
	 *
	 * \tparam T the type of the value to retrieve
	 *
	 * \return value retrieved from the variant
	 *
	 * \throws bad_cast exception if the variant stores different type than specified
	 */
	template < typename F, typename std::enable_if < std::is_same < F, T >::value >::type * = nullptr >
	const F && get ( ) const & {
		return std::move ( m_value );
	}

	/**
	 * \brief
	 * Allows to retrieve a value from the variant.
	 *
	 * \tparam T the type of the value to retrieve
	 *
	 * \return value retrieved from the variant
	 *
	 * \throws bad_cast exception if the variant stores different type than specified
	 */
	template < typename F, typename std::enable_if < std::is_same < F, T >::value >::type * = nullptr >
	F && get ( ) && {
		return std::move ( m_value );
	}

	/**
	 * \brief
	 * Callback executor on current variant value.
	 *
	 * \tparam Result the uniform result type
	 * \tparam Callable type of callable able to accept any variant type.
	 *
	 * \param callable object of callable type able to accept any variant type via function call operator. The possible void type is represented by call with no parameters
	 *
	 * \return the result of the callable on the current variable type
	 */
	template < class Result, class Callable >
	Result call ( Callable callable ) const {
		return callable ( m_value );
	}

	/**
	 * \brief
	 * Operator to print the variant to the output stream.
	 *
	 * \param out the output stream
	 * \param obj the variant to print
	 *
	 * \return the output stream from the \p out
	 */
	friend std::ostream & operator << ( std::ostream & out, const variant < T > & obj ) {
		return out << obj.m_value;
	}

	/**
	 * \brief
	 * To string cast operator.
	 *
	 * \return string representation of the variant
	 */
	explicit operator std::string ( ) const {
		return ( std::string ) m_value;
	}

	/**
	 * \brief
	 * Allows to create the variant from defaultly constructed value of specified type.
	 *
	 * \tparam T the type to default construct ańd assign to newly constructed variant
	 *
	 * \return new variant with content of defaultly constructed value of type T
	 */
	static variant < T > from ( ) {
		return variant < T > ( T { } );
	}
};

/**
 * \brief
 * Specialisation of the compare structure implementing the three-way comparison.
 *
 * \tparam Ts ... types in the variant
 */
template<typename ... Ts>
struct compare < ext::variant < Ts ... > > {

	/**
	 * \brief
	 * Implementation of the three-way comparison.
	 *
	 * \param first the left operand of the comparison
	 * \param second the right operand of the comparison
	 *
	 * \return negative value of left < right, positive value if left > right, zero if left == right
	 */
	int operator()(const ext::variant < Ts ... > & first, const ext::variant < Ts ... > & second) const {
		return first.compare(second);
	}
};

/**
 * \brief
 * Overload of to_string function.
 *
 * \param value the vector to be converted to string
 *
 * \tparam Ts ... types in the variant
 *
 * \return string representation
 */
template < class ... Ts >
std::string to_string ( const ext::variant < Ts ... > & value ) {
	return ( std::string ) value;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator < ( const ext::variant < Ts ... > & first, const T & second ) {
	return first.compare ( second ) < 0;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator > ( const ext::variant < Ts ... > & first, const T & second ) {
	return first.compare ( second ) > 0;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator <= ( const ext::variant < Ts ... > & first, const T & second ) {
	return first.compare ( second ) <= 0;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator >= ( const ext::variant < Ts ... > & first, const T & second ) {
	return first.compare ( second ) >= 0;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator == ( const ext::variant < Ts ... > & first, const T & second ) {
	return first.compare ( second ) == 0;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator != ( const ext::variant < Ts ... > & first, const T & second ) {
	return first.compare ( second ) != 0;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator < ( const T & first, const ext::variant < Ts ... > & second ) {
	return second > first;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator > ( const T & first, const ext::variant < Ts ... > & second ) {
	return second < first;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator <= ( const T & first, const ext::variant < Ts ... > & second ) {
	return second >= first;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator >= ( const T & first, const ext::variant < Ts ... > & second ) {
	return second <= first;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator == ( const T & first, const ext::variant < Ts ... > & second ) {
	return second == first;
}

template < class ... Ts, class T, typename std::enable_if < ext::is_in < T, Ts ... >::value >::type * = nullptr >
bool operator != ( const T & first, const ext::variant < Ts ... > & second ) {
	return second != first;
}

/**
 * \brief
 * Class to help building of the variant type or, in case variant is requested to be constructed from single type or more types but all the same, that concrete type.
 *
 * \tparam Res the resulting contstructed variant
 * \tparam Ts ... pack of types to construct the variant from
 */
template < class Res, class ... Ts >
struct variant_builder_impl {
};

/**
 * \brief
 * Specialisation of the helper class to build the variant type as the type T as case when the variant is requested to be constructed from single type.
 *
 * \tparam T type to construct the variant from
 */
template < class T >
struct variant_builder_impl < ext::variant < >, T > {
	/**
	 * \brief
	 * The type T as simplification of the variant to the type T.
	 */
	typedef ext::variant < T > type;
};

/**
 * \brief
 * Recursion ending specialisation of the helper class to build the variant type. The last type T is added as last type to the variant parameters.
 *
 * \tparam ResTs ... pack of types already accepted as types of the variant
 * \tparam T last type to construct the variant from
 */
template < class ... ResTs, class T >
struct variant_builder_impl < ext::variant < ResTs ... >, T > {
	/**
	 * \brief
	 * The resulting variant type.
	 */
	typedef ext::variant < ResTs ..., T > type;
};

/**
 * \brief
 * Recursive step of class to build the variant type.
 *
 * \tparam ResTs ... pack of types already accepted as types of the variant
 * \tparam T next type to construct the variant from
 * \tparam Ts ... remaining types to add to variant template parameters
 */
template < class ... ResTs, class T, class ... Ts >
struct variant_builder_impl < ext::variant < ResTs ... >, T, Ts ... > {
	/**
	 * The type T is tested whether it is included in the Ts ... pack. If it is included it is skipped, if not it is added to types of resulting variant.
	 */
	typedef typename variant_builder_impl < typename std::conditional < is_in < T, Ts ... >::value, ext::variant < ResTs ... >, ext::variant < ResTs ..., T > >::type, Ts ... >::type type;
};

/**
 * \brief
 * Class to construct variant from types pack. The construction drops duplicate types from the Ts ... pack. Variant of single type is converted to that type.
 *
 * \tparam Ts ... pack of types to build the variant for
 */
template < class ... Ts >
using variant_builder = typename variant_builder_impl < ext::variant < >, Ts ... >::type;

} /* namespace ext */

#endif /* __VARIANT_HPP_ */
